-- Carousel by Astromech
-- Contact on WoWInterface

-- TODO
-- Option to ignore launchers
-- After showing a recently updated plugin, cycle back to the previous plugin after a certain amount of time.


local Carousel = LibStub("AceAddon-3.0"):NewAddon("Carousel", "AceConsole-3.0");
local objects = {}
local bars = {}
local objectPositions = {}
Carousel.objects = objects
Carousel.bars = bars
Carousel.objectPositions = objectPositions

local defaults = {
	profile = {
		version = nil,
		notify = true,
		frameList = {"1"},
		frames = {
			["*"] = {
				scale = 1,
				frameX = 0,
				frameY = 0,
				objectOrder = {},
				height = 24,
				widthOffset = 16,
				iconLeft = true,
				textAlignment = "LEFT",
				blockPos = "CENTER",
				anchorSide = "CENTER",
				hideArrows = false,
				backdropColor = {0, 0, 0, .3},
				backdropBorderColor = {0, 0, 0, .7},
				fontsize = 12,
				fontoutline = nil,
				textColor = {1, 1, 1, 1},
				labelColor = {1, 1, 0, 1},
				borderSize = 16,
				borderInsets = 5,
				font = "Fonts\\FRIZQT__.ttf",
				LSMFontKey = nil,
				space = 8,
				arrowSpace = 4,
				width = 0,
				locked = true,
				combatHideTooltips = nil,
			},
		},
		objectDBs = {
			["*"] = {
				showLabel = true,
				showText = true,
				showIcon = true,
				autoshow = false,
			},
		},
		foundObjects = {},
	},
}

function Carousel:OnInitialize()
	Carousel.db = LibStub("AceDB-3.0"):New("CarouselDB", defaults, "Default")
	Carousel.db.RegisterCallback(Carousel, "OnProfileChanged", "OnProfileChanged")
	Carousel.db.RegisterCallback(Carousel, "OnProfileReset", "OnProfileChanged")
	Carousel.db.RegisterCallback(Carousel, "OnProfileCopied", "OnProfileChanged")
	local f = CreateFrame("Frame", nil, InterfaceOptionsFrame)
	f:SetScript("OnShow", function(self)
		self:SetScript("OnShow", nil)
		LoadAddOn("CarouselOptions")
	end)
end

function Carousel:OnEnable() 
	Carousel:InitializeObjects()
	Carousel:InitializeCommands()
	Carousel:OnProfileChanged()
end

function Carousel:CheckVersion()
	local old = Carousel.db.profile.version or 0
	local db = Carousel.db.profile
	if old < 20000 then
		-- Previously had a pre-20000 profile. Migrate relevant settings to new database structure
		db.frameList = {"1"}
		local migrate = {"backdropColor", "backdropBorderColor", "scale",
			"fontsize", "fontoutline", "textColor", "labelColor", "height", "iconLeft", "textAlignment",
			"blockPos", "borderSize", "borderInsets", "font", "LSMFontKey"}
		for key in pairs(migrate) do
			if db[key] ~= nil then db.frames["1"][key] = db[key] end
			db[key] = nil
		end
		local erase = {"lock", "arrowBackdropColor", "arrowBackdropBorderColor", "arrowColor", "width",
			"anchor", "point", "frameX", "frameY"}
		for key in pairs(erase) do
			db[key] = nil
		end
		for obj in pairs(db.objectDBs) do
			db.objectDBs[obj].enabled = nil
			db.objectDBs[obj].autoresize = nil
		end
		local order = db.objectOrder or {}
		db.objectOrder = nil
		db.frames["1"].objectOrder = {[1] = {}}
		for i,name in pairs(order) do
			db.frames["1"].objectOrder[1][i] = name
		end
	end
	if old < 21000 then
		for _,frame in pairs(db.frameList) do
			db.frames[frame].frameX = db.frames[frame].frameX * db.frames[frame].scale
			db.frames[frame].frameY = db.frames[frame].frameY * db.frames[frame].scale
		end
	end
	-- if next version
		-- more conversions... etc
	local curVersion = 21010
	db.version = curVersion
	if old < curVersion then
		if (Carousel.db.profile.notify) then Carousel:Print("Upgraded profile to version to "..curVersion) end
	elseif old > curVersion then
		if (Carousel.db.profile.notify) then Carousel:Print("Profile's version was higher than the current version. "..
			"This could result in unexpected behavior. Please update to the newest version of Carousel.") end
	end
end

do
	local cache = {}
	function Carousel:InitializeFrames()
		local frames = {}
		for i,name in pairs(Carousel.db.profile.frameList) do
			frames[name] = true
		end
		for name,bar in pairs(bars) do
			if not frames[name] then
				bar:Dispose()
				cache[bar] = true
				bars[name] = nil
			end
		end
		for name in pairs(frames) do
			if not bars[name] then
				local bar = next(cache)
				if bar then cache[bar] = nil end
				bars[name] = Carousel:CreateNewFrame(name, bar)
				bars[name]:Update()
			end
		end
	end
end

function Carousel:DeleteBar(name)
	local list = Carousel.db.profile.frameList
	for i,n in pairs(list) do
		if n == name then
			tremove(list, i)
			break
		end
	end
	local order = Carousel.db.profile.frames[name].objectOrder
	for k in pairs(order) do
		order[k] = nil
	end
	Carousel:OnProfileChanged()
end

function Carousel:AddBar(name)
	for i,bar in pairs(Carousel.db.profile.frameList) do
		if bar == name then
			return
		end
	end
	tinsert(Carousel.db.profile.frameList, name)
	Carousel:OnProfileChanged()
end

function Carousel:RemakeObjectLocations()
	for name,tab in pairs(objectPositions) do
		tab.bar = nil
		tab.dock = nil
		tab.place = nil
	end
	for i,bar in pairs(Carousel.db.profile.frameList) do
		for dock,order in pairs(Carousel.db.profile.frames[bar].objectOrder) do
			for place,name in pairs(order) do
				local tab = objectPositions[name] or {}
				objectPositions[name] = tab
				tab.bar = bar
				tab.dock = dock
				tab.place = place
			end
		end
	end
end

function Carousel:InitializeObjects()
	Carousel:RemakeObjectLocations()
	local LDB = LibStub("LibDataBroker-1.1")
	for name, obj in LDB:DataObjectIterator() do
		Carousel:DataObjectCreated(nil, name, obj)
	end
	LDB.RegisterCallback(Carousel, "LibDataBroker_DataObjectCreated", "DataObjectCreated")
	LDB.RegisterCallback(Carousel, "LibDataBroker_AttributeChanged__text", "UpdateText")
	LDB.RegisterCallback(Carousel, "LibDataBroker_AttributeChanged__label", "UpdateText")
	LDB.RegisterCallback(Carousel, "LibDataBroker_AttributeChanged__value", "UpdateText")
	LDB.RegisterCallback(Carousel, "LibDataBroker_AttributeChanged__suffix", "UpdateText")
	LDB.RegisterCallback(Carousel, "LibDataBroker_AttributeChanged__icon", "UpdateIcon")
	LDB.RegisterCallback(Carousel, "LibDataBroker_AttributeChanged__iconR", "UpdateIcon")
	LDB.RegisterCallback(Carousel, "LibDataBroker_AttributeChanged__iconG", "UpdateIcon")
	LDB.RegisterCallback(Carousel, "LibDataBroker_AttributeChanged__iconB", "UpdateIcon")
	LDB.RegisterCallback(Carousel, "LibDataBroker_AttributeChanged__iconCoords", "UpdateIcon")
end

do
	local AceCfgCmd
	function ChatCmd(input)
		if Carousel.Options or LoadAddOn("CarouselOptions") then
			AceCfgCmd = AceCfgCmd or LibStub("AceConfigCmd-3.0")
			if not input or input:trim() == "" then
				InterfaceOptionsFrame_OpenToCategory("Carousel")
			else
				AceCfgCmd.HandleCommand(Carousel, "carousel", "Carousel", input)
			end
		else
			Carousel:Print("CarouselOptions not enabled.")
		end
	end
	function Carousel:InitializeCommands()
		Carousel:RegisterChatCommand("carousel", ChatCmd)
		Carousel:RegisterChatCommand("car", ChatCmd)
	end
end

function Carousel:OnProfileChanged()
	Carousel:CheckVersion()
	Carousel:RemakeObjectLocations()
	Carousel:InitializeFrames()
	for _, frame in pairs(bars) do
		frame:Update()
	end
	if Carousel.Options then Carousel:UpdateOptions() end
end

local supportedTypes = {["data source"] = true, ["launcher"] = true}
function Carousel:DataObjectCreated(_, name, obj)
	if not supportedTypes[obj.type] then return end
	objects[name] = obj
	if not objectPositions[name] then
		objectPositions[name] = {}
	end
	local found = Carousel.db.profile.foundObjects
	if not found[name] then
		found[name] = true
		if (Carousel.db.profile.notify) then Carousel:Print("New LDB feed, "..name.." found! "..
			"Type \"/carousel\" to open the config menu and assign it to a Carousel.") end
	end
	if objectPositions[name].bar and bars[objectPositions[name].bar] then
		local dock = bars[objectPositions[name].bar].docks[objectPositions[name].dock]
		if objectPositions[name].place == 1 or dock.active == nil then
			dock:SetActiveObject(name)
		end
	end
	if Carousel.Options then Carousel:RegisterObjectOptions(name) end
end	

function Carousel:VarifyFont(bar)
	Carousel.db.profile.frames[bar].font = defaults.profile.frames["*"].font
end

function Carousel:AddObject(name, destbar, destdock)
	local pos = objectPositions[name]
	assert(pos.bar == nil)
	local dock = bars[destbar].docks[destdock]
	table.insert(dock.order, name)
	pos.bar = destbar
	pos.dock = destdock
	pos.place = #dock.order
	if not dock.active then 
		dock:ResetActive() 
	end
	local db = Carousel.db.profile.frames[destbar]
	dock:UpdateLayout(db.hideArrows, db.blockPos, db.arrowSpace)
end

function Carousel:RemoveObject(name, destbar, destdock)
	local pos = objectPositions[name]
	assert(pos.bar == destbar and pos.dock == destdock)
	local dock = bars[destbar].docks[destdock]
	table.remove(dock.order, pos.place)
	pos.bar = nil
	pos.dock = nil
	pos.place = nil
	for i,obj in pairs(dock.order) do
		objectPositions[obj].place = i
	end
	if dock.active == name then
		dock:CycleRight()
	end
	local db = Carousel.db.profile.frames[destbar]
	dock:UpdateLayout(db.hideArrows, db.blockPos, db.arrowSpace)
end

function Carousel:SwitchObject(oldname, newname, destbar, destdock)
	local opos = objectPositions[oldname]
	local npos = objectPositions[newname]
	local dock = bars[destbar].docks[destdock]
	assert(opos.bar == destbar and opos.dock == destdock)
	assert(npos.bar == nil and npos.dock == nil)
	local place = opos.place
	assert(dock.order[place] == oldname)
	opos.bar = nil
	opos.dock = nil
	opos.place = nil
	npos.bar = destbar
	npos.dock = destdock
	npos.place = place
	dock.order[place] = newname
	if dock.active == oldname then
		dock:CycleRight()
	end
end	

function Carousel:UpdateText(_, name)
	if not objects[name] then return end
	local pos = objectPositions[name]
	if pos.bar then
		bars[pos.bar].docks[pos.dock]:UpdateText(name)
	end
end

function Carousel:UpdateIcon(_, name, obj)
	if not objects[name] then return end
	local pos = objectPositions[name]
	if pos.bar then
		bars[pos.bar].docks[pos.dock]:UpdateIcon(name)
	end
end
