
local Carousel = LibStub("AceAddon-3.0"):GetAddon("Carousel")
if not Carousel then return end

local DockProto = CreateFrame"Frame"
Carousel.DockProto = DockProto
local meta = {__index = DockProto}
local tinsert, tremove = table.insert, table.remove

function Carousel:CreateNewDock(index, bar)
	local frame = setmetatable(CreateFrame("Frame", nil, Carousel.bars[bar]), meta)
	frame.bar = bar
	frame.index = index
	frame.display = CreateFrame("Button", nil, frame)
	frame.display:EnableMouse(true)
	--frame.display:SetClampedToScreen(true)
	frame.display:EnableMouseWheel(true)
	frame.display:RegisterForClicks("LeftButtonUp", "RightButtonUp")
	frame.display:SetScript("OnClick", DockProto.OnClick)
	frame.display:SetScript("OnEnter", DockProto.OnEnter)
	frame.display:SetScript("OnLeave", DockProto.OnLeave)
	frame.display:SetScript("OnMouseWheel", DockProto.OnMouseWheel)
	frame.text = frame.display:CreateFontString(nil, "OVERLAY")
	frame.text:SetFont("Fonts\\FRIZQT__.TTF", 11)
	frame.icon = frame.display:CreateTexture(nil, "OVERLAY")
	frame:Enable(index, bar)
	return frame
end

function DockProto:Enable(index, bar)
	self.index = index
	self.bar = bar
	self:SetParent(Carousel.bars[bar])
	self:Show()
	self.idealWidth = 0
	self.arrowWidth = 0
	self.displayWidth = 0
	return self
end

function DockProto:Disable()
	self.index = nil
	self.bar = nil
	self:Hide()
	self:SetParent(nil)
	self:SetActiveObject(nil)
	self:ClearAllPoints()
	return self
end

function DockProto:CreateArrows(index)
	local db = Carousel.db.profile.frames[self.bar]
	self.leftArrow = CreateFrame("Button", nil, self)
	self.leftArrow:EnableMouse(true)
	self.leftArrow:RegisterForClicks("LeftButtonUp")
	self.leftArrow:SetScript("OnClick", self.LeftArrowOnClick)
	self.rightArrow = CreateFrame("Button", nil, self)
	self.rightArrow:EnableMouse(true)
	self.rightArrow:RegisterForClicks("LeftButtonUp")
	self.rightArrow:SetScript("OnClick", self.RightArrowOnClick)
	self.leftText = self.leftArrow:CreateFontString(nil, "OVERLAY")
	self.rightText = self.rightArrow:CreateFontString(nil, "OVERLAY")
	self.leftText:SetFont(db.font, db.fontsize, db.fontoutline)
	self.rightText:SetFont(db.font, db.fontsize, db.fontoutline)
	local label = self.labelColor or "|cffffffff"
	self.leftText:SetFormattedText("|c%s<|r", label)
	self.rightText:SetFormattedText("|c%s>|r", label)
	self.leftText:SetAllPoints(self.leftArrow)
	self.rightText:SetAllPoints(self.rightArrow)
	self:UpdateHeight()
end

function DockProto:UpdateHeight(height)
	height = height or Carousel.db.profile.frames[self.bar].height
	self:SetHeight(height)
	self.icon:SetWidth(height/1.5)
	self.icon:SetHeight(height/1.5)
	if self.leftArrow then
		self.leftArrow:SetWidth(height)
		self.rightArrow:SetWidth(height)
	end
end

function DockProto:UpdateLayout(hideArrows, blockPos, space)
	self.display:ClearAllPoints()
	self.display:SetPoint("TOP", self)
	self.display:SetPoint("BOTTOM", self)
	self.arrowsShown = (not hideArrows and #self.order > 1)
	if self.arrowsShown and not self.leftArrow then
		self:CreateArrows()
	end
	if not self.leftArrow then
		self.display:SetPoint("LEFT", self)
		self.display:SetPoint("RIGHT", self)
	else
		self.leftArrow:ClearAllPoints()
		self.rightArrow:ClearAllPoints()
		self.leftArrow:SetPoint("TOP", self)
		self.leftArrow:SetPoint("BOTTOM", self)
		self.rightArrow:SetPoint("TOP", self)
		self.rightArrow:SetPoint("BOTTOM", self)
		self.display:SetPoint("TOP", self)
		self.display:SetPoint("BOTTOM", self)
		if not self.arrowsShown then
			self.leftArrow:Hide()
			self.rightArrow:Hide()
			self.display:SetPoint("LEFT", self)
			self.display:SetPoint("RIGHT", self)
		elseif blockPos == "CENTER" then
			self.leftArrow:Show()
			self.rightArrow:Show()
			self.leftArrow:SetPoint("LEFT", self)
			self.rightArrow:SetPoint("RIGHT", self)
			self.display:SetPoint("LEFT", self.leftArrow, "RIGHT", space, 0)
			self.display:SetPoint("RIGHT", self.rightArrow, "LEFT", -space, 0)
		elseif blockPos == "RIGHT" then
			self.leftArrow:Show()
			self.rightArrow:Show()
			self.leftArrow:SetPoint("LEFT", self)
			self.rightArrow:SetPoint("LEFT", self.leftArrow, "RIGHT", space, 0)
			self.display:SetPoint("LEFT", self.rightArrow, "RIGHT", space, 0)
			self.display:SetPoint("RIGHT", self)
		else
			self.leftArrow:Show()
			self.rightArrow:Show()
			self.display:SetPoint("LEFT", self)
			self.rightArrow:SetPoint("RIGHT", self)
			self.leftArrow:SetPoint("RIGHT", self.rightArrow, "LEFT", -space, 0)
			self.display:SetPoint("RIGHT", self.leftArrow, "LEFT", -space, 0)
		end
	end
	self:ResizeToText()
end

function DockProto:UpdateTextIconPosition(left)
	if not self.bar then return end
	left = left or Carousel.db.profile.frames[self:GetParent().name].iconLeft
	self.text:ClearAllPoints()
	self.icon:ClearAllPoints()
	if not self.icon:IsShown() then
		self.text:SetPoint("LEFT", self.display)
		self.text:SetPoint("RIGHT", self.display)
	elseif left then
		self.icon:SetPoint("LEFT", self.display)
		self.text:SetPoint("LEFT", self.icon, "RIGHT", 4, 0)
		self.text:SetPoint("RIGHT", self.display)
	else
		self.text:SetPoint("LEFT", self.display)
		self.icon:SetPoint("RIGHT", self.display)
		self.text:SetPoint("RIGHT", self.icon, "RIGHT", -4, 0)
	end
end

function DockProto:UpdateFont(...)
	self.text:SetFont(...)
	if self.leftArrow then
		self.leftText:SetFont(...)
		self.rightText:SetFont(...)
	end
end

function DockProto:UpdateTextColor(text, label)
	self.textColor = text
	self.labelColor = label
	if self.leftArrow then
		self.leftText:SetFormattedText("|c%s<|r", label)
		self.rightText:SetFormattedText("|c%s>|r", label)
	end
	self:UpdateText(self.active)
end

function DockProto:UpdateTextAlignment(align)
	self.text:SetJustifyH(align)
end

function DockProto:UpdateOrder(order)
	self.order = order
	self:ResetActive()
end

function DockProto:CycleRight()
	local offset = (self.active and Carousel.objectPositions[self.active].place) or 0
	local max = #self.order
	local index, name
	for i=1,max do
		index = i + offset
		index = index > max and index - max or index
		name = self.order[index]
		if Carousel.objects[name] then
			self:SetActiveObject(name)
			return
		end
	end
	self:SetActiveObject(nil)
end

function DockProto:CycleLeft()
	local offset = (self.active and Carousel.objectPositions[self.active].place) or 2
	local max = #self.order
	local index, name
	for i=1,max do
		index = offset - i
		index = index <= 0 and max + index or index
		name = self.order[index]
		if Carousel.objects[name] then
			self:SetActiveObject(name)
			return
		end
	end
	self:SetActiveObject(nil)
end
	
function DockProto:SetActiveObject(name)
	if self.active then
		local oldobj = Carousel.objects[self.active]
		if oldobj.OnHide then oldobj.OnHide(self.display) end
	end
	self.active = name
	if name then
		local newobj = Carousel.objects[self.active]
		if newobj.OnShow then newobj.OnShow(self.display) end
	end
	self:UpdateText(name)
	self:UpdateIcon(name)
end

function DockProto:ResetActive()
	self:SetActiveObject(nil)
	self:CycleRight()
end

function DockProto:UpdateText(name)
	if name == nil then
		self.text:SetFormattedText("|c%sCarousel|r", self.labelColor or "|cffffffff")
		self:ResizeToText()
	elseif self.active ~= name then
		local pos = Carousel.objectPositions[name]
		if pos.bar == self.bar and pos.dock == self.index and Carousel.db.profile.objectDBs[name].autoshow then
			self:SetActiveObject(name)
		end
	else
		self:SetText(name)
	end
end

function DockProto:UpdateIcon(name)
	if name == nil then
		local shown = self.icon:IsShown()
		self.icon:Hide()
		if shown then self:UpdateTextIconPosition() end
	elseif self.active ~= name then
		local pos = Carousel.objectPositions[name]
		if pos.bar == self.bar and pos.dock == self.index and Carousel.db.profile.objectDBs[name].autoshow then
			self:SetActiveObject(name)
		end
	else
		self:SetIcon(name)
	end
end

function DockProto:SetText(name)
	local obj = Carousel.objects[name]
	local db = Carousel.db.profile.objectDBs[name]
	local textColor, labelColor = self.textColor or "|cffffffff", self.labelColor or "|cffffffff"
	if obj.type == "launcher" then
		if db.showText then
			local addonName, title = GetAddOnInfo(obj.tocname or name)
			self.text:SetFormattedText("|c%s%s|r", textColor, title or addonName or name)
		end
	elseif obj.type == "data source" then
		if db.showLabel and db.showText then
			if obj.value and obj.suffix then
				self.text:SetFormattedText("|c%s%s:|r |c%s%s|r|c%s%s|r", 
					labelColor, obj.label or name, textColor, tostring(obj.value), labelColor, obj.suffix)
			else
				self.text:SetFormattedText("|c%s%s:|r |c%s%s|r", 
					labelColor, obj.label or name, textColor, obj.text or "")
			end
		elseif db.showLabel then
			self.text:SetFormattedText("|c%s%s|r", labelColor, obj.label or name)
		elseif db.showText then
			if obj.value and obj.suffix then
				self.text:SetFormattedText("|c%s%s|r|c%s%s|r", 
					textColor, tostring(obj.value), labelColor, obj.suffix or "")
			else
				self.text:SetFormattedText("|c%s%s|r", textColor, obj.text or "")
			end
		else
			self.text:SetText()
		end
	end
	self:ResizeToText()
end

function DockProto:SetIcon(name)
	local obj = Carousel.objects[name]
	local db = Carousel.db.profile.objectDBs[name]
	local shown = self.icon:IsShown()
	if db.showIcon and obj.icon then
		self.icon:SetTexture(obj.icon)
		if obj.iconR and obj.iconG and obj.iconB then
			self.icon:SetVertexColor(obj.iconR, obj.iconG, obj.iconR)
		else
			self.icon:SetVertexColor(1, 1, 1, 1)
		end
		if obj.iconCoords then
			self.icon:SetTexCoord(unpack(obj.iconCoords))
		else
			self.icon:SetTexCoord(0, 1, 0, 1)
		end
		self.icon:Show()
	else
		self.icon:Hide()
	end
	if shown ~= self.icon:IsShown() then
		self:UpdateTextIconPosition()
		self:ResizeToText()
	end
end

function DockProto:ResizeToText()
	if not self.bar then return end
	local bar = Carousel.bars[self.bar]
	local width = self.text:GetStringWidth()
	if self.icon:IsShown() then
		width = width + self.icon:GetWidth() + 4
	end
	self.displayWidth = width
	if self.arrowsShown then
		self.arrowWidth = (self:GetHeight() + Carousel.db.profile.frames[self.bar].arrowSpace) * 2
	else
		self.arrowWidth = 0
	end
	self.idealWidth = self.arrowWidth + self.displayWidth
	if Carousel.db.profile.frames[self.bar].width == 0 then
		local oldwidth = self:GetWidth()
		self.display:SetWidth(self.displayWidth)
		self:SetWidth(self.idealWidth)
		bar:SetWidth(bar:GetWidth() - oldwidth + self.idealWidth)
	else
		bar:UpdateWidth()
	end
end

function DockProto.LeftArrowOnClick(arrow)
	arrow:GetParent():CycleLeft()
end

function DockProto.RightArrowOnClick(arrow)
	arrow:GetParent():CycleRight()
end

function DockProto.OnClick(display, ...)
	local active = display:GetParent().active
	if not active then return end
	local obj = Carousel.objects[active]
	if obj and obj.OnClick then
		obj.OnClick(display, ...)
	end
end

do
	local function getAnchors(frame)
		local _, y = frame:GetCenter()
		if y < GetScreenHeight() / 2 then
			return "BOTTOM", "TOP"
		else
			return "TOP", "BOTTOM"
		end
	end

	local function GT_OnLeave(self)
		self:SetScript("OnLeave", self.CarouselOldOnLeave)
		if self.CarouselOldOnLeave then self.CarouselOldOnLeave(self) end
		self:EnableMouse(self.CarouselOldMouse)
		self.CarouselOldOnLeave = nil
		self.CarouselOldMouse = nil
		self.CarouselActivated = nil
		self:Hide()
	end

	local function PrepareTooltip(frame, parent)
		if frame.CarouselActivated then return end
		frame.CarouselActivated = true
		frame.CarouselOldOnLeave = frame:GetScript("OnLeave")
		frame.CarouselOldMouse = frame:IsMouseEnabled()
		frame:EnableMouse(true)
		frame:SetScript("OnLeave", GT_OnLeave)
		frame:SetOwner(parent, "ANCHOR_NONE")
		frame:ClearAllPoints()
		local a1, a2 = getAnchors(parent)
		frame:SetPoint(a1, parent, a2)	
	end

	function DockProto.OnEnter(display, ...)
		if Carousel.db.profile.frames[display:GetParent().bar].combatHideTooltips and InCombatLockdown() then return end
		local active = display:GetParent().active
		if not active then return end
		local obj = Carousel.objects[active]
		if obj.OnTooltipShow then
			PrepareTooltip(GameTooltip, display)
			obj.OnTooltipShow(GameTooltip)
			GameTooltip:Show()
		elseif obj.tooltip then
			PrepareTooltip(obj.tooltip, display)
			obj.tooltip:Show()
		end
		if obj.OnEnter then
			obj.OnEnter(display, ...)
		end
	end

	function DockProto.OnLeave(display, ...)
		local active = display:GetParent().active
		if not active then return end
		local obj = Carousel.objects[active]
		if not MouseIsOver(GameTooltip) and obj.OnTooltipShow then
			GT_OnLeave(GameTooltip)
		end
		if obj.tooltip and not MouseIsOver(obj.tooltip) then
			GT_OnLeave(obj.tooltip)
		end
		if obj.OnLeave then
			obj.OnLeave(display, ...)
		end
	end
end

function DockProto.OnMouseWheel(display, dir)
	local frame = display:GetParent()
	if frame.active then frame.OnLeave(display) end
	if dir == 1 then
		frame:CycleRight()
	else
		frame:CycleLeft()
	end
	if frame.active then frame.OnEnter(display) end
end
