
local Carousel = LibStub("AceAddon-3.0"):GetAddon("Carousel")
if not Carousel then return end

local L = LibStub("AceLocale-3.0"):GetLocale("Carousel")
local feedMapping = {}
local barMapping = {}

local function frameGetFactory(level, mod)
	return function(info)
		local depth = #info
		if mod then
			return mod(Carousel.db.profile.frames[barMapping[info[depth-level]]][info[depth]])
		else
			return Carousel.db.profile.frames[barMapping[info[depth-level]]][info[depth]]
		end
	end
end
local function frameSetFactory(level, update, mod)
	return function(info, value)
		local depth = #info
		local name = barMapping[info[depth-level]]
		if mod then value = mod(value) end
		Carousel.db.profile.frames[name][info[depth]] = value
		if update then
			Carousel.bars[name][update](Carousel.bars[name])
		end
	end
end
local function feedGetFactory(level)
	return function(info)
		local depth = #info
		return Carousel.db.profile.objectDBs[feedMapping[info[depth-level]]][info[depth]]
	end
end
local function feedSetFactory(level, update)
	return function(info, value)
		local depth = #info
		local name = feedMapping[info[depth-level]]
		Carousel.db.profile.objectDBs[name][info[depth]] = value
		if update then
			Carousel[update](Carousel, nil, name)
		end
	end
end
local function frameColorGetFactory(level)
	return function(info)
		local depth = #info
		local c = Carousel.db.profile.frames[barMapping[info[depth-level]]][info[depth]]
		return c[1], c[2], c[3], c[4] or 1
	end
end
local function frameColorSetFactory(level, update)
	return function(info, r, g, b, a)
		local depth = #info
		local name = barMapping[info[depth-level]]
		local c = Carousel.db.profile.frames[name][info[depth]]
		c[1] = r
		c[2] = g
		c[3] = b
		c[4] = a or 1
		if update then
			Carousel.bars[name][update](Carousel.bars[name])
		end
	end
end
	
local function copy(tab)
	local into = {}
	for k,v in pairs(tab) do
		if type(v) == "table" then
			into[k] = copy(v)
		else
			into[k] = v
		end
	end
	return into
end

local Options = {}
Options.core = {
	type = "group",
	name = "Carousel",
	childGroups = "tab",
	args = {
		general = {
			name = L.GeneralTab,
			type = "group",
			order = 1,
			args = {
				author = {
					type = "description",
					order = 1,
					name = "Astromech",
				},
				contact = {
					type = "description",
					order = 2,
					name = "WoWInterface",
				},
				description = {
					type = "description",
					order = 3,
					name = L.CarouselDescription,
				},
				addtab = {
					type = "input",
					order = 4,
					name = L.BarCreate,
					desc = L.BarCreateDesc,
					get = function() return "" end,
					set = function(info, value)
						Carousel:AddBar(string.gsub(tostring(value), " ", ""))
					end,
				},
				notify = {
					type = "toggle",
					order = 5,
					name = L.Notifications,
					desc = L.NotificationsDesc,
					get = function() return Carousel.db.profile.notify end,
					set = function(_, val) Carousel.db.profile.notify = val end,
				},
			},
		},
		feeds = {
			name = L.DataFeedsTab,
			type = "group",
			childGroups = "tree",
			order = 3,
			get = feedGetFactory(1),
			args = {},
		},	
	},
}
Options.core.args.profiles = LibStub("AceDBOptions-3.0"):GetOptionsTable(Carousel.db)
Options.core.args.profiles.order = 2
Options.feedTemplate = {
	type = "group",
	order = 0,
	args = {
		showText = {
			name = L.FeedShowText,
			type = "toggle",
			order = 0,
			set = feedSetFactory(1, "UpdateText"),
		},
		showLabel = {
			name = L.FeedShowLabel,
			type = "toggle",
			order = 1,
			set = feedSetFactory(1, "UpdateText"),
		},
		showIcon = {
			name = L.FeedShowIcon,
			type = "toggle",
			order = 2,
			set = feedSetFactory(1, "UpdateIcon"),
		},
		autoshow = {
			name = L.FeedAutoshow,
			desc = L.FeedAutoshowDesc,
			type = "toggle",
			order = 3,
			set = feedSetFactory(1, nil),
		},
	},
}

function Carousel:RegisterObjectOptions(name)
	local tabname = gsub(name, " ", "_")
	feedMapping[tabname] = name
	local otab = copy(Options.feedTemplate)
	otab.name = name
	-- Support object specific options provided by the data feed?
	Options.core.args.feeds.args[tabname] = otab
	if not Carousel.objectPositions[name].bar then
		Options.unallocatedFeeds[name] = name
	end
end

Options.barLookTemplate = {
	type = "group",
	childGroups = "tree",
	get = frameGetFactory(2),
	args = {
		possize = {
			name = L.BarPositionAndSize,
			order = 0,
			type = "group",
			args = {
				scale = {
					name = L.BarScale,
					order = 0,
					type = "range",
					min = .64, max = 2, step = .01, bigStep = .1,
					set = frameSetFactory(2, "UpdateScale"),
				},
				frameX = {
					name = L.BarPosX,
					order = 1,
					type = "input",
					get = frameGetFactory(2, tostring),
					set = frameSetFactory(2, "UpdateAnchor", tonumber),
				},
				frameY = {
					name = L.BarPosY,
					order = 2,
					type = "input",
					get = frameGetFactory(2, tostring),
					set = frameSetFactory(2, "UpdateAnchor", tonumber),
				},
				locked = {
					name = L.BarLock,
					order = 3,
					type = "toggle",
					desc = L.BarLockDesc,
					set = frameSetFactory(2, "UpdateLock"),
				},
				anchorSide = {
					name = L.BarAnchorSide,
					order = 4,
					type = "select",
					values = {["LEFT"] = L.Left, ["RIGHT"] = L.Right, ["CENTER"] = L.Center},
					set = frameSetFactory(2, "UpdateAnchor"),
				},
				height = {
					name = L.BarHeight,
					order = 5,
					type = "range",
					min = 20, max = 60, step = 1,
					set = frameSetFactory(2, "UpdateHeight"),
				},
				width = {
					name = L.BarWidth,
					order = 6,
					type = "input",
					desc = L.BarWidthDesc,
					get = frameGetFactory(2, tostring),
					set = frameSetFactory(2, "UpdateFixedWidth", tonumber),
				},
				widthOffset = {
					name = L.BarWidthOffset,
					order = 7,
					type = "range",
					min = 0, max = 32, step = 1,
					set = frameSetFactory(2, "UpdateSpace"),
				},
				space = {
					name = L.BarDockSpace,
					order = 8,
					type = "range",
					min = 0, max = 20, step = 1,
					set = frameSetFactory(2, "UpdateSpace"),
				},
			},
		},
		arrows = {
			name = L.BarArrows,
			order = 1,
			type = "group",
			set = frameSetFactory(2, "UpdateLayout"),
			args = {
				hideArrows = {
					name = L.BarHideArrows,
					order = 0,
					type = "toggle",
				},
				blockPos = {
					name = L.BarBlockPos,
					desc = L.BarBlockPosDesc,
					order = 1,
					type = "select",
					values = {["LEFT"] = L.Left, ["RIGHT"] = L.Right, ["CENTER"] = L.Center},
				},
				arrowSpace = {
					name = L.BarArrowSpace,
					order = 2,
					type = "range",
					min = 0, max = 20, step = 1,
				},
			},
		},
		textpos = {
			name = L.BarTextPos,
			order = 2,
			type = "group",
			args = {
				iconLeft = {
					name = L.BarIconPos,
					order = 0,
					type = "select",
					values = {["LEFT"] = L.Left, ["RIGHT"] = L.Right},
					get = frameGetFactory(2, function(val) return val and "LEFT" or "RIGHT" end),
					set = frameSetFactory(2, "UpdateTextIconPosition", function(val) return val == "LEFT" end),
				},
				textAlignment = {
					name = L.BarTextAlignment,
					order = 1,
					type = "select",
					values = {["LEFT"] = L.Left, ["RIGHT"] = L.Right, ["CENTER"] = L.Center},
					set = frameSetFactory(2, "UpdateTextAlignment"),
				},
			},
		},
		backdrop = {
			name = L.BarBackdrop,
			order = 3,
			type = "group",
			args = {
				backdropColor = {
					name = L.BarBackdropColor,
					order = 0,
					type = "color",
					hasAlpha = true,
					get = frameColorGetFactory(2),
					set = frameColorSetFactory(2, "UpdateBackdrop"),
				},
				backdropBorderColor = {
					name = L.BarBackdropBorderColor,
					order = 1,
					type = "color",
					hasAlpha = true,
					get = frameColorGetFactory(2),
					set = frameColorSetFactory(2, "UpdateBackdrop"),
				},
				borderSize = {
					name = L.BarBorderSize,
					order = 2,
					type = "range",
					min = 0, max = 32, step = 1,
					set = frameSetFactory(2, "UpdateBackdrop"),
				},
				borderInsets = {
					name = L.BarBorderInsets,
					order = 3,
					type = "range",
					min = -16, max = 16, step = 1,
					set = frameSetFactory(2, "UpdateBackdrop"),
				},
			},
		},
		font = {
			name = L.BarFont,
			order = 4,
			type = "group",
			args = {
				-- font reserved for other addon
				fontsize = {
					name = L.BarFontSize,
					order = 0,
					type = "range",
					min = 8, max = 32, step = 1,
					set = frameSetFactory(2, "UpdateFont"),
				},
				fontoutline = {
					name = L.BarFontOutline,
					order = 1,
					type = "select",
					values = {["NONE"] = L.None, ["OUTLINE"] = L.Outline, ["THICKOUTLINE"] = L.ThickOutline, ["MONOCHROME"] = L.Monochrome},
					get = frameGetFactory(2, function(val)
						if val == nil then
							return "NONE"
						else
							return val
						end
					end),
					set = frameSetFactory(2, "UpdateFont", function(val) 
						if val == "NONE" then 
							return nil 
						else 
							return val 
						end
					end),
				},
				textColor = {
					name = L.BarTextColor,
					order = 2,
					type = "color",
					get = frameColorGetFactory(2),
					set = frameColorSetFactory(2, "UpdateTextColor"),
				},
				labelColor = {
					name = L.BarLabelColor,
					order = 3,
					type = "color",
					get = frameColorGetFactory(2),
					set = frameColorSetFactory(2, "UpdateTextColor"),
				},
			},
		},
		tooltip = {
			name = L.BarTooltip,
			order = 5,
			type = "group",
			args = {
				combatHideTooltips = {
					name = L.BarTooltipCombatHide,
					order = 0,
					type = "toggle",
					set = frameSetFactory(2, nil),
				},
			},
		},
	},
}
	
Options.barOrderTemplate = {
	type = "group",
	args = {
		delete = {
			name = L.BarDelete,
			desc = L.BarDeleteDesc,
			order = 0,
			type = "execute",
			func = function(info) 
				local name = barMapping[info[#info-1]]
				Carousel:DeleteBar(barMapping[info[#info-1]])
			end,
		},
		create = {
			name = L.BarNewDock,
			desc = L.BarNewDockDesc,
			order = -1,
			type = "execute",
			func = function(info)
				Carousel:AddDockOptions(barMapping[info[#info-1]])
			end,
		},
	}
}

Options.barOrderDockTemplate = {
	type = "group",
	inline = true,
	args = {
		delete = {
			name = L.BarDeleteDock,
			desc = L.BarDeleteDockDesc,
			order = 0,
			type = "execute",
			func = function(info)
				local depth = #info
				Carousel:RemoveDockOptions(barMapping[info[depth-2]], tonumber(info[depth-1]))
			end,
		},
		newFeed = {
			name = L.BarAddFeedToDock,
			desc = L.BarAddFeedToDockDesc,
			order = -1,
			type = "select",
			values = function()
				local tab = copy(Options.unallocatedFeeds)
				tab[""] = ""
				return tab
			end,
			get = function() return "" end,
			set = function(info, val) 
				if val == "" then return end
				local depth = #info
				Carousel:AddDockFeedOptionsEntry(barMapping[info[depth-2]], tonumber(info[depth-1]), val)
			end,
		},
	},
}

Options.barOrderDockFeedTemplate = {
	name = L.BarFeedEntry,
	desc = L.BarFeedEntryDesc,
	type = "select",
	values = function(info) 
		local tab = copy(Options.unallocatedFeeds)
		tab[info.arg] = info.arg
		tab["!REMOVE!"] = L.BarFeedEntryRemove
		return tab
	end,
	get = function(info) return info.arg end,
	set = function(info, val)
		if val == info.arg then return end
		local depth = #info
		if val == "!REMOVE!" then
			Carousel:RemoveDockFeedOptionsEntry(barMapping[info[depth-2]], tonumber(info[depth-1]), info.arg)
		else
			Carousel:ChangeDockFeedOptionsEntry(barMapping[info[depth-2]], tonumber(info[depth-1]), info.arg, val)
		end
	end,
}

Options.unallocatedFeeds = {}
Options.bars = {}

do
	local count = Options.core.args.feeds.order
	function Carousel:RegisterFrameOptions(name)
		Options.bars[name] = true
		local oltab = copy(Options.barLookTemplate)
		local ootab = copy(Options.barOrderTemplate)
		oltab.name = L.BarTabLook:format(name) --"Bar: "..name.." Look"
		oltab.order = count + 1
		ootab.name = L.BarTabFeeds:format(name) --"Bar: "..name.." Feeds"
		ootab.order = count + 2
		count = count + 2
		Options.core.args[name.."-look"] = oltab
		Options.core.args[name.."-order"] = ootab
		barMapping[name.."-look"] = name
		barMapping[name.."-order"] = name
	end
end

function Carousel:RemoveFrameOptions(name)
	Options.bars[name] = nil
	Options.core.args[name.."-look"] = nil
	Options.core.args[name.."-order"] = nil
	barMapping[name.."-look"] = nil
	barMapping[name.."-order"] = nil
end

function Carousel:UpdateFrameOptions()
	local found = {}
	for i,frame in pairs(Carousel.db.profile.frameList) do
		if not Options.bars[frame] then
			Carousel:RegisterFrameOptions(frame)
		end
		found[frame] = true
	end
	for frame in pairs(Options.bars) do
		if not found[frame] then
			Carousel:RemoveFrameOptions(frame)
		end
	end
	for obj in pairs(Options.unallocatedFeeds) do
		if Carousel.objectPositions[obj] then
			Options.unallocatedFeeds[obj] = nil
		else
			Options.unallocatedFeeds[obj] = obj
		end
	end
end			

function Carousel:AddDockOptions(bar, index)
	index = index or Carousel.bars[bar]:AddDock()
	local otab = copy(Options.barOrderDockTemplate)
	otab.name = L.BarDockName:format(index)
	otab.order = index
	Options.core.args[bar.."-order"].args[tostring(index)] = otab
end

function Carousel:RemoveDockOptions(bar, index)
	local removedFeeds = Carousel.bars[bar]:RemoveDock(index)
	oorder = Options.core.args[bar.."-order"].args
	local i = index
	while oorder[tostring(i)] do
		local stri = tostring(i)
		oorder[stri] = oorder[tostring(i+1)]
		if oorder[stri] then
			oorder[stri].name = L.BarDockName:format(i)
			oorder[stri].order = i
		end
		i = i + 1
	end
	for feed in pairs(removedFeeds) do
		Options.unallocatedFeeds[feed] = feed
	end
end

function Carousel:AddDockFeedOptionsEntry(bar, dock, feed, ignore)
	if not ignore then Carousel:AddObject(feed, bar, dock) end
	Options.unallocatedFeeds[feed] = nil
	local otab = copy(Options.barOrderDockFeedTemplate)
	otab.order = Carousel.objectPositions[feed].place
	otab.arg = feed
	Options.core.args[bar.."-order"].args[tostring(dock)].args[tostring(Carousel.objectPositions[feed].place)] = otab
end

function Carousel:ChangeDockFeedOptionsEntry(bar, dock, oldfeed, feed)
	local place = Carousel.objectPositions[oldfeed].place
	Carousel:SwitchObject(oldfeed, feed, bar, dock)
	Options.unallocatedFeeds[oldfeed] = oldfeed
	Options.unallocatedFeeds[feed] = nil
	Options.core.args[bar.."-order"].args[tostring(dock)].args[tostring(place)].arg = feed
end
	
function Carousel:RemoveDockFeedOptionsEntry(bar, dock, feed)
	local place = Carousel.objectPositions[feed].place
	Carousel:RemoveObject(feed, bar, dock)
	if Carousel.objects[feed] then
		Options.unallocatedFeeds[feed] = feed
	end
	local ops = Options.core.args[bar.."-order"].args[tostring(dock)].args
	local i = place
	while ops[tostring(i)] do
		local stri = tostring(i)
		ops[stri] = ops[tostring(i+1)]
		if ops[stri] then ops[stri].order = i end
		i = i + 1
	end
end	

function Carousel:UpdateObjectOrderOptions()
	for name,tab in pairs(Carousel.objectPositions) do
		if tab.bar then
			Options.unallocatedFeeds[name] = nil
		else
			Options.unallocatedFeeds[name] = name
		end
	end
	for n,bar in pairs(Carousel.db.profile.frameList) do
		local dockopts = Options.core.args[bar.."-order"].args
		for i = 1,#dockopts do
			dockopts[tostring(i)] = nil
		end
		for dock,order in pairs(Carousel.db.profile.frames[bar].objectOrder) do
			Carousel:AddDockOptions(bar, dock)
			for place,feed in pairs(order) do
				Carousel:AddDockFeedOptionsEntry(bar, dock, feed, true)
			end
		end
	end
end

function Carousel:UpdateOptions()
	Carousel:UpdateFrameOptions()
	Carousel:UpdateObjectOrderOptions()
end

for name in pairs(Carousel.objects) do
	Carousel:RegisterObjectOptions(name)
end

Options.barMapping = barMapping
Options.feedMapping = feedMapping
Carousel.Options = Options
if Carousel.OnOptionsLoaded then Carousel:OnOptionsLoaded() end
Carousel:UpdateOptions()
LibStub("AceConfigRegistry-3.0"):RegisterOptionsTable("Carousel", Options.core)
LibStub("AceConfigDialog-3.0"):AddToBlizOptions("Carousel", "Carousel")
